import jax.numpy as jnp
import jax


def poisson_log_partition(x):
    """
    Poisson link from the exponential family.
    See Wikipedia 'Exponential families'.

    Parameters
    ----------
    x: jax.numpy.ndarray

    Returns
    -------
    jax.numpy.ndarray
    """
    return jnp.exp(x)


def binomial_log_partition(x, n):
    """
    Binomial link from the exponential family.
    See Wikipedia 'Exponential families'.
    Parameters
    ----------
    x: jax.numpy.ndarray
    n:int

    Returns
    -------
    jax.numpy.ndarray
    """
    return n * jax.numpy.logaddexp(jnp.array([0]), x)



import torch
import numpy as np

from scud.pca import _PCA
from scud.logpartitions import poisson_log_partition
from scud.lognormalizers import poisson_log_normalizer
from scud._initialisation import _PoissonReg, _init_components_poisson_model


class PlnPCA(_PCA):
    """
    Poisson Log-Normal Principal Component Analysis. The log
    partitions is the exponential, the link function is the PCA while
    the log normalizer is the log factorial.
    """

    def log_partition(self, x):
        """Exponential log partition."""
        return poisson_log_partition(x)

    def log_normalizer(self, x):
        """Log normalizer of a poisson law: log factorial."""
        return poisson_log_normalizer(x)

    def _get_init_params(self):
        poissreg = _PoissonReg()
        poissreg.fit(
            torch.from_numpy(np.asarray(self._data)),
            torch.from_numpy(np.asarray(self.exog)).float(),
            torch.from_numpy(np.asarray(self.offsets)),
        )
        coef = poissreg.beta.detach().numpy()
        components = _init_components_poisson_model(
            torch.from_numpy(np.asarray(self._data)), self.latent_dimension
        )
        return {"coef": coef, "components": components}



import jax.numpy as jnp

from scud.logpartitions import binomial_log_partition
from scud.lognormalizers import binomial_log_normalizer
from scud.pca import _PCA
from scud._initialisation import (
    _MBinomialRegression,
    _init_components_binomial_model,
)


class BinPCA(_PCA):
    """
    Binomial Principal Component Analysis. The log
    partitions is n*log(1 + exp(x)), the link function is the PCA while
    the log normalizer is log of binomial coefficients.
    """

    _n_max: int

    def log_partition(self, x):
        """Log partition that is n*log(1+ exp(x))."""
        return binomial_log_partition(x, self._n_max)

    def log_normalizer(self, x):
        """Log normalizer of a binomial law: log binomial coefficient."""
        return binomial_log_normalizer(x, self._n_max)

    def _init_arguments(self):
        super()._init_arguments()
        self._n_max = jnp.max(self._data)

    def _get_init_params(self):
        """
        Intialisation for the coefficients is done with a binomial regression.
        Intialisation for the components is done by a PCA on the logit of the normalized counts.
        """
        binreg = _MBinomialRegression(self._n_max)
        binreg.fit(
            self._data,
            self.exog,
            self.offsets,
        )
        coef = binreg.coef
        components = _init_components_binomial_model(
            self._data, self.latent_dimension, self._n_max
        )
        return {"coef": coef, "components": components}



import math
from typing import Union
import textwrap


import jax.numpy as jnp
import jax.numpy.linalg as nla
from jax import vmap
import torch
import numpy as np
import pandas as pd


def _log_gaussian_density(x: jnp.ndarray, mean: jnp.ndarray, cov_matrix: jnp.ndarray):
    dim = x.shape[-1]
    sign, logdet = nla.slogdet(cov_matrix)
    log_const = dim / 2 * math.log(2 * math.pi) + 1 / 2 * sign * logdet
    x_moins_mean = x - mean
    inv_sig = jnp.linalg.inv(cov_matrix)
    log_density_unnormalized = (
        -1 / 2 * jnp.matmul(jnp.matmul(inv_sig, x_moins_mean), x_moins_mean)
    )
    return log_density_unnormalized - log_const


def _log_unit_gaussian_density(x):
    dim = x.shape[-1]
    log_const = dim / 2 * math.log(2 * math.pi)
    log_density_unnormalized = -1 / 2 * jnp.sum(x**2, axis=-1)
    return -log_const + log_density_unnormalized


def _vmap_log_gaussian_density(
    x: jnp.ndarray, mean: jnp.ndarray, cov_matrix: jnp.ndarray
):
    return vmap(_log_gaussian_density, in_axes=(0, 0, 0))(x, mean, cov_matrix)


def _vvmap_log_gaussian_density(
    x: jnp.ndarray, mean: jnp.ndarray, cov_matrix: jnp.ndarray
):
    return vmap(_vmap_log_gaussian_density, in_axes=(0, None, None))(
        x, mean, cov_matrix
    )


def _log_binom(n, endog):
    _endog = endog + (endog == 0) - (endog == n)
    return n * jnp.log(n) - _endog * jnp.log(_endog) - (n - _endog) * jnp.log(_endog)


def _log_stirling(integer: jnp.ndarray) -> jnp.ndarray:
    """
    Compute log(n!) using the Stirling formula.

    Parameters
    ----------
    integer : jnp.ndarray
        Input tensor

    Returns
    -------
    jnp.ndarray
        Approximation of log(n!) element-wise.
    """
    integer_ = integer + (integer == 0)  # Replace 0 with 1 since 0! = 1!
    return jnp.log(jnp.sqrt(2 * jnp.pi * integer_)) + integer_ * jnp.log(
        integer_ / math.exp(1)
    )


def _format_data(
    data: Union[torch.Tensor, jnp.ndarray, pd.DataFrame, np.ndarray]
) -> torch.Tensor or None:
    """
    Transforms the data in a jnp.ndarray if the input is an array, and None if the input is None.
    Raises an error if the input is not an array or None.

    Parameters
    ----------
    data : pd.DataFrame, np.ndarray, torch.Tensor, jnp.ndarray
        Input data.

    Returns
    -------
    jnp.ndarray or None
        Formatted data.

    Raises
    ------
    AttributeError
        If the value is not an array or None.
    """
    if data is None:
        return None
    if isinstance(data, pd.DataFrame):
        return jnp.array(data.values)
    if isinstance(data, (np.ndarray, jnp.ndarray)):
        return data
    if isinstance(data, torch.Tensor):
        return data.numpy()
    raise AttributeError(
        "Please insert either a numpy.ndarray, pandas.DataFrame, torch.Tensor or jnp.ndarray."
    )


def _format_dict_of_array(dict_array):
    for array in dict_array.values():
        array = _format_data(array)
    return dict_array


def _add_doc(parent_class, *, params=None, example=None, returns=None, see_also=None):
    def wrapper(fun):
        doc = getattr(parent_class, fun.__name__).__doc__
        if doc is None:
            doc = ""
        doc = textwrap.dedent(doc).rstrip(" \n\r")
        if params is not None:
            doc += textwrap.dedent(params.rstrip(" \n\r"))
        if returns is not None:
            doc += "\n\nReturns"
            doc += "\n-------"
            doc += textwrap.dedent(returns)
        if see_also is not None:
            doc += "\n\nSee also"
            doc += "\n--------"
            doc += textwrap.dedent(see_also)
        if example is not None:
            doc += "\n\nExamples"
            doc += "\n--------"
            doc += textwrap.dedent(example)
        fun.__doc__ = doc
        return fun

    return wrapper


def _sample_gaussians(number_of_samples, mean, sqrt_variance):
    """
    Sample some gaussians with the right mean and variance.
    Careful, we ask for the square root of Sigma, not Sigma.

    Args:
         number_of_samples : int. the number of samples wanted.
         mean : jnp.ndarray of size (batch_size,latent_dim)
         sqrt_variance : jnp.ndarray or size
            (batch_size, latent_dim, latent_dim). The square roots matrices
             of the covariance matrices. (e.g. if you want to sample a
             gaussian with covariance matrix A, you need to give the
            square root of A in argument.)

    Returns:
        gaussians: jnp.ndarray of size
        (number_of_samples, batch_size,latent_dim). It is a vector
        of number_of_samples gaussian of dimension mean.shape. For each
        1< i< NSample, 1<k< nBatch , W[i,k] is a gaussian with mean mean[k,:]
        and variance sqrt_variance[k,:,:]@sqrt_variance[k,:,:].
    """
    latent_dim = mean.shape[1]
    centered_gaussian = np.random.randn(number_of_samples, 1, latent_dim, 1)
    gaussians = jnp.matmul(
        jnp.expand_dims(sqrt_variance, 0), centered_gaussian
    ).squeeze() + jnp.expand_dims(mean, 0)
    return gaussians


def _vmap_all_but_params(func, x, params, additional_data):
    return vmap(
        func,
        in_axes=(
            0,
            {key: None for key in params.keys()},
            {key: 0 for key in additional_data.keys()},
        ),
    )(x, params, additional_data)


def _vmap_only_x(func, x, params, additional_data):
    return vmap(
        func,
        in_axes=(
            0,
            {key: None for key in params.keys()},
            {key: None for key in additional_data.keys()},
        ),
    )(x, params, additional_data)



import importlib.metadata

from .base import BaseModel
from .logpartitions import poisson_log_partition, binomial_log_partition
from .linkfunctions import linear, linear_with_offsets
from .lognormalizers import poisson_log_normalizer
from .plnpca import PlnPCA
from .binpca import BinPCA
from .inverse_parameter_mappings import poisson_mapping, binomial_mapping
from .pca_sampling_parameters import (
    get_components,
    get_exog,
    get_coef,
    get_offsets,
    get_linear_params_and_additional_data,
)
from .sample import PlnPCAsampler, BinomialPCAsampler

__version__ = importlib.metadata.version("scud")



from scud._utils import _log_stirling


def poisson_log_normalizer(x):
    """The log normalize constant in the poisson case.

    Parameters
    ----------
    x: jax.numpy.ndarray

    Returns
    -------
    jax.numpy.ndarray
    """
    return _log_stirling(x)


def binomial_log_normalizer(x, n):
    """Log normalize constant in the binomial case.
    Parameters
    ----------
    x: jax.numpy.ndarray
    n: int

    Returns
    -------
    jax.numpy.ndarray
    """
    num = _log_stirling(n)
    denom_left = _log_stirling(x)
    denom_right = _log_stirling(n - x)
    return num - denom_left - denom_right



import jax.numpy as np
from jax import random

COMPONENT_KEY = "components"
COV_KEY = "exog"
COEF_KEY = "coef"
OFFSETS_KEY = "offsets"


def get_components(
    dim: int = 50, latent_dimension: int = 10, seed: int = 0
) -> np.ndarray:
    """
    Generate components for a linear model.

    Parameters
    ----------
    dim : int, optional
        The total number of features in the data, by default 50.
    latent_dimension : int, optional
        The number of latent dimensions, by default 10.
    seed : int, optional
        Random seed for reproducibility, by default 0.

    Returns
    -------
    components : jax.numpy.ndarray
        The generated components for the linear model.
    """
    key = random.PRNGKey(seed)
    block_size = dim // latent_dimension
    components = np.zeros((dim, latent_dimension))

    for column_number in range(latent_dimension):
        start_idx = column_number * block_size
        end_idx = (column_number + 1) * block_size
        components = components.at[start_idx:end_idx, column_number].set(1)

    components += random.normal(key=key, shape=(dim, latent_dimension)) / 8
    return components


def get_coef(nb_cov: int = 1, dim: int = 50, seed: int = 0) -> np.ndarray:
    """
    Generate coefficients.

    Parameters
    ----------
    nb_cov : int, optional
        The number of exog, by default 1.
    dim : int, optional
        The total number of features in the data, by default 50.
    seed : int, optional
        Random seed for reproducibility, by default 0.

    Returns
    -------
    coefficients : jax.numpy.ndarray
        The generated coefficients for linear regression.
    """
    key = random.PRNGKey(seed)
    coefficients = random.normal(key=key, shape=(nb_cov, dim))
    return coefficients


def get_exog(n_samples: int = 100, nb_cov: int = 1, seed: int = 0) -> np.ndarray:
    """
    Generate binary exog.

    Parameters
    ----------
    n_samples : int, optional
        The number of samples, by default 100.
    nb_cov : int, optional
        The number of exog, by default 1.
    seed : int, optional
        Random seed for reproducibility, by default 0.

    Returns
    -------
    exog : jax.numpy.ndarray
        The generated binary exog.
    """
    key = random.PRNGKey(seed)
    exog = random.bernoulli(key=key, shape=(n_samples, nb_cov))
    return exog


def get_offsets(
    n_samples: int = 100, dim: int = 50, seed: int = 0, only_zero: bool = False
) -> np.ndarray:
    """
    Generate offsets for linear regression.

    Parameters
    ----------
    n_samples : int, optional
        The number of samples, by default 100.
    dim : int, optional
        The total number of features in the data, by default 50.
    seed : int, optional
        Random seed for reproducibility, by default 0.
    only_zero : bool, optional
        If True, only zero offsets are generated, by default False.

    Returns
    -------
    offsets : jax.numpy.ndarray
        The generated offsets for linear regression.
    """
    key = random.PRNGKey(seed)
    offsets = random.normal(key=key, shape=(n_samples, dim))
    return offsets * 0 if only_zero else offsets


def get_linear_params_and_additional_data(  # pylint: disable=too-many-arguments
    n_samples: int = 100,
    nb_cov: int = 1,
    dim: int = 50,
    latent_dimension: int = 10,
    no_offsets: bool = False,
    seed: int = 0,
) -> dict:
    """
    Generate linear regression parameters and additional data (offsets and exog).

    Parameters
    ----------
    n_samples : int, optional
        The number of samples, by default 100.
    nb_cov : int, optional
        The number of exog, by default 1.
    dim : int, optional
        The total number of features in the data, by default 50.
    latent_dimension : int, optional
        The number of latent dimensions, by default 10.
    no_offsets : bool, optional
        If True, generate data without offsets, by default False.
    seed : int, optional
        Random seed for reproducibility, by default 0.

    Returns
    -------
    dictionary : dict
        A dictionary containing components, coefficients, exog, and offsets.
    """
    components = get_components(dim, latent_dimension, seed=seed)
    coefficients = get_coef(nb_cov, dim, seed=seed)
    exog = get_exog(n_samples, nb_cov, seed=seed)
    offsets = get_offsets(n_samples, dim, seed=seed, only_zero=no_offsets)

    return {
        COMPONENT_KEY: components,
        COEF_KEY: coefficients,
        COV_KEY: exog,
        OFFSETS_KEY: offsets,
    }



from typing import Dict

import jax
import jax.numpy as np


def linear_with_offsets(
    x: jax.numpy.ndarray,
    *,
    params: Dict[str, jax.numpy.ndarray],
    additional_data: Dict[str, jax.numpy.ndarray],
) -> jax.numpy.ndarray:
    """
    Linear link function with offsets. Performs Principal
    Component Analysis (PCA) and add offsets.

    Parameters
    ----------
    x : jax.numpy.ndarray
        Input values.
    params : dict
        Model parameters including components and coef.
    additional_data : dict
        Additional data including offsets and exog.

    Returns
    -------
    jax.numpy.ndarray
        Output values from the linear link function with offsets.
    """
    linear_result = linear(x, params=params, additional_data=additional_data)
    return linear_result + additional_data["offsets"]


def linear(
    x: jax.numpy.ndarray,
    *,
    params: Dict[str, jax.numpy.ndarray],
    additional_data: Dict[str, jax.numpy.ndarray],
) -> jax.numpy.ndarray:
    """
    Linear link function. Performs Principal Component Analysis (PCA).

    Parameters
    ----------
    x : jax.numpy.ndarray
        Input values.
    params : dict
        Model parameters including components and coef.
    additional_data : dict
        Additional data including exog and coef.

    Returns
    -------
    jax.numpy.ndarray
        Output values from the linear link function.
    """
    components_x = np.matmul(params["components"], x)
    mean = np.matmul(additional_data["exog"], params["coef"])
    return components_x + mean



import numpy as np

from scud.base import BaseModel
from scud.linkfunctions import linear_with_offsets


class _PCA(BaseModel):
    _nb_exog = None

    def _init_arguments(self):
        super()._init_arguments()
        if self._additional_data is None:
            self._nb_exog = 1
        else:
            self._nb_exog = self._additional_data["exog"].shape[1]

    @property
    def nb_exog(self):
        """Number of exogenous variable in the model."""
        return self._nb_exog

    @property
    def coef(self):
        """The coefficient in the regression."""
        return self._params["coef"]

    @property
    def components(self):
        """The components in the PCA."""
        return self._params["components"]

    @property
    def covariance(self):
        """The covariance of the model."""
        return self.components @ (self.components.T)

    def _get_init_params(self):
        coef = np.random.rand(self.nb_exog, self.dim)
        components = np.random.randn(self.dim, self.latent_dimension)
        return {"coef": coef, "components": components}

    def link_function(self, x, params, additional_data):
        """
        Linear link function with offsets.
        """
        return linear_with_offsets(x, params=params, additional_data=additional_data)



import jax.numpy as jnp


def _get_weights_and_const_from_log_quotient(log_num, log_denom):
    diff_log = log_num - log_denom
    const = jnp.max(diff_log, axis=0)
    diff_log -= const
    weights = jnp.exp(diff_log)
    return weights, const


def _normalize_weights(weights):
    return jnp.divide(weights, jnp.sum(weights, axis=0))


def _get_normalized_weights_const_and_weights(log_num, log_denom):
    weights, const = _get_weights_and_const_from_log_quotient(log_num, log_denom)
    return _normalize_weights(weights), const, weights



import jax.numpy as np
import jax


def poisson_mapping(x):
    """Mapping to simulate from a poisson law. See Wikipédia 'Exponential Families'.
    Parameters
    ----------
    x: jax.numpy.ndarray

    Returns
    -------
    jax.numpy.ndarray

    """
    return np.exp(x)


def binomial_mapping(x):
    """Mapping to simulate from a binomial law. See Wikipédia 'Exponential Families'.
    Parameters
    ----------
    x: jax.numpy.ndarray

    Returns
    -------
    jax.numpy.ndarray
    """
    return jax.nn.sigmoid(x)



from abc import ABC, abstractmethod
from typing import Optional, Dict, Union
from math import sqrt

import math
import torch
import optax
import numpy as np
import pandas as pd
import jax.numpy as jnp
from jax import vmap, grad, hessian

from scud._utils import (
    _format_dict_of_array,
    _format_data,
    _sample_gaussians,
    _vvmap_log_gaussian_density,
    _log_unit_gaussian_density,
    _vmap_all_but_params,
    _vmap_only_x,
)
from scud._weights import _get_normalized_weights_const_and_weights

DELTA = 0.1


class BaseModel(ABC):
    """Implements all the features that can be shared across models."""

    def __init__(
        self,
        data: Union[torch.Tensor, np.ndarray, pd.DataFrame, jnp.ndarray],
        *,
        latent_dimension: int,
        additional_data: Optional[
            Dict[str, Union[torch.Tensor, np.ndarray, pd.DataFrame, jnp.ndarray]]
        ] = None,
        alpha: float = 0.1,
    ):
        """
        Init the model.
        Parameters
        ----------
        data: torch.Tensor or np.ndarray or pd.DataFrame or jnp.ndarray
            The input array on which doing dimension reduction.
        latent_dimension: int (keyword-only)
            The reduced dimension.
        additional_data:dict[str,Union[torch.Tensor,np.ndarray,pd.DataFrame,jnp.ndarray]](kw-only)
            Optional dictionnary with additional data such as covariates or offsets. Each item
            should be either a torch.Tensor or np.ndarray or pd.DataFrame or jnp.ndarray.
            Default to None.
        alpha : float
            The parameter of the mixture between the best proposal and the
            defensive proposal. Should be between 0 and 1. Default to 0.1
        Returns
        -------
        An instance of the BaseModel class. Cannot be instantiated.
        """
        self._data = _format_data(data)
        self._aux_var = {
            "latent_dimension": latent_dimension,
            "nb_particles": None,
            "loglike_list": [],
            "selected_indices": None,
            "batch_size": None,
            "alpha": None,
        }
        self._additional_data = _format_dict_of_array(additional_data)
        self._init_arguments()
        self._params = self._get_init_params()
        self._gaussian_params = {
            "mean_prop": None,
            "var_prop": None,
            "sqrt_var_prop": None,
        }
        self._particles = None
        self._aux_var["alpha"] = alpha

    @property
    def exog(self):
        """The exogenous variable in the model."""
        return self._additional_data["exog"]

    @property
    def offsets(self):
        """The offsets in the model."""
        return self._additional_data["offsets"]

    @property
    def n_samples(self):
        """the number of samples in the dataset (number of lines of the given data)"""
        return self._aux_var["n_samples"]

    @property
    def dim(self):
        """Number of variables of data."""
        return self._aux_var["dim"]

    @property
    def _selected_indices(self):
        return self._aux_var["selected_indices"]

    @property
    def _alpha(self):
        return self._aux_var["alpha"]

    @property
    def _batch_size(self):
        return self._aux_var["batch_size"]

    @property
    def _current_batch_size(self):
        return len(self._selected_indices)

    @property
    def _nb_batch(self):
        return self._aux_var["nb_batch"]

    @property
    def latent_dimension(self):
        """The latent dimension of the model"""
        return self._aux_var["latent_dimension"]

    @property
    def nb_particles(self):
        """Number of particles used for the Monte Carlo approximation."""
        return self._aux_var["nb_particles"]

    @property
    def particles(self):
        """The particles used for Monte Carlo approximation."""
        return self._particles

    @property
    def _batch_particles(self):
        return self._particles[:, self._selected_indices]

    @property
    def _batch_data(self):
        return self._data[self._selected_indices]

    @property
    def _batch_additional_data(self):
        batch_dict = {}
        for key, items in self._additional_data.items():
            batch_dict[key] = items[self._selected_indices]
        return batch_dict

    @property
    def mean_prop(self):
        """The mean of the proposal."""
        return self._gaussian_params["mean_prop"]

    @property
    def _batch_mean_prop(self):
        return self.mean_prop[self._selected_indices]

    @property
    def var_prop(self):
        """The variance of the proposal."""
        return self._gaussian_params["var_prop"]

    @property
    def _sqrt_var_prop(self):
        """The variance of the proposal."""
        return self._gaussian_params["sqrt_var_prop"]

    @property
    def _batch_var_prop(self):
        return self.var_prop[self._selected_indices]

    @property
    def _batch_sqrt_var_prop(self):
        return self._sqrt_var_prop[self._selected_indices]

    def _vmap_link_function(
        self, x: jnp.ndarray, params: dict, additional_data: dict
    ) -> jnp.ndarray:
        """
        Vectorized map of the link function.

        Parameters
        ----------
        x : jax.numpy.ndarray
            Input values.
        params : dict
            Model parameters.
        additional_data : dict
            Additional data such as covariates or offsets.

        Returns
        -------
        jax.numpy.ndarray
            Vectorized output values from the link function.
        """
        return _vmap_all_but_params(self.link_function, x, params, additional_data)

    def _vvmap_link_function(
        self, x: jnp.ndarray, *, params: dict, additional_data: dict
    ) -> jnp.ndarray:
        """
        Doubly Vectorized vectorized map of the link function.

        Parameters
        ----------
        x : jax.numpy.ndarray
            Input values.
        params : dict (keyword-only)
            Model parameters.
        additional_data : dict (keyword-only)
            Additional data.

        Returns
        -------
        jax.numpy.ndarray
            Doubly vectorized output values from the link function.
        """
        return _vmap_only_x(self._vmap_link_function, x, params, additional_data)

    def _hessian_w_complete_loglike(self, data, x, params, additional_data):
        return hessian(self._complete_loglike, argnums=1)(
            data, x, params, additional_data
        )

    def _vmap_hessian_w_complete_loglike(self, data, x, params, additional_data):
        return vmap(
            self._hessian_w_complete_loglike,
            in_axes=(
                0,
                0,
                {key: None for key in params.keys()},
                {key: 0 for key in additional_data.keys()},
            ),
        )(data, x, params, additional_data)

    def _complete_loglike(self, data, x, params, additional_data):
        """Computes the complete log likelihood of the model."""
        return self._log_conditional_y_w(
            data, x, params, additional_data
        ) + _log_unit_gaussian_density(x)

    def _vmap_complete_loglike(self, data, x, params, additional_data):
        """
        Computes the complete log likelihood of the model vectorized
        one time.
        """
        parameter = self._vmap_link_function(
            x, params=params, additional_data=additional_data
        )
        log_unit_gauss = _log_unit_gaussian_density(x)
        return self._log_nef(data, parameter) + log_unit_gauss

    def _log_nef(self, data, parameter):
        """log density of the Natural Exponential Family."""
        log_partition_parameter = self.log_partition(parameter)
        return jnp.sum(
            data * parameter - log_partition_parameter - self.log_normalizer(data),
            axis=-1,
        )

    def _log_conditional_y_w(self, data, x, params, additional_data):
        """Log conditional of Y given W"""
        parameter = self.link_function(
            x, params=params, additional_data=additional_data
        )
        return jnp.sum(self._log_nef(data, parameter))

    def _grad_theta_log_conditional_y_w(self, data, x, params, additional_data):
        """Derivative with respect to theta of the conditional."""
        return grad(self._log_conditional_y_w, argnums=2)(
            data, x, params, additional_data
        )

    def _vmap_grad_theta_log_conditional_y_w(self, data, x, params, additional_data):
        """Derivative with respect to theta of the log conditional Y given W vectorized."""
        return vmap(
            self._grad_theta_log_conditional_y_w,
            in_axes=(
                0,
                0,
                {key: None for key in params.keys()},
                {key: 0 for key in additional_data.keys()},
            ),
        )(data, x, params, additional_data)

    def _vvmap_grad_theta_log_conditional_y_w(self, data, x, params, additional_data):
        """Doubly vectorized derivate with respect to theta of the log conditional Y given W"""
        return vmap(
            self._vmap_grad_theta_log_conditional_y_w,
            in_axes=(
                None,
                0,
                {key: None for key in params.keys()},
                {key: None for key in additional_data.keys()},
            ),
        )(data, x, params, additional_data)

    def _grad_w_complete_loglike(self, data, x, params, additional_data):
        return grad(self._complete_loglike, argnums=1)(data, x, params, additional_data)

    def _vvmap_conditional(self, data, x, params, additional_data):
        parameter = self._vvmap_link_function(
            x, params=params, additional_data=additional_data
        )
        return self._log_nef(data, parameter)

    def _vvmap_complete_loglike(self, data, x, params, additional_data):
        """Computes the doubly vectorized complete log likelihood of the model."""
        return self._vvmap_conditional(
            data, x, params, additional_data
        ) + _log_unit_gaussian_density(x)

    def transform(self):
        """Transforms data."""
        return self.mean_prop

    def _get_loglike(self, weights, const):
        per_sample_loglike = jnp.log(jnp.mean(weights, axis=0)) + const
        return jnp.mean(per_sample_loglike, axis=0)

    def _snis_estimator(self, grad_params, normalized_weights):
        normalized_weights = jnp.expand_dims(normalized_weights, (2, 3))
        for key in grad_params.keys():
            grad_params[key] *= normalized_weights
            grad_params[key] = -jnp.sum(grad_params[key], axis=(0, 1))
        return grad_params

    def fit(
        self,
        *,
        nb_particles: int = 30,
        nb_max_gradients: int = -1,
        batch_size: int = None,
        nb_epoch: int = 100,
    ):
        """
        Fits the data with gradient ascent via self-normalized estimated gradients.

        Parameters
        ----------
        nb_particles : int (optional, keyword-only)
            The number of particles used for Monte Carlo
            approximation. default to 30.

        nb_max_gradients: int (optional, keyword-only)
            The maximum number of gradients to compute.
            One epoch computes n_samples gradients. If -1,
            will do only nb_epoch and not look at the number of gradients computed.
            Default to -1.

        batch_size: int (optional, keyword-only)
            The batch size used. If None, will take the number of samples.
            Default to None.

        nb_epoch: int (optional, keyword-only)
            The number of epoch to perform. If nb_max_gradients is lower
            than n_samples * nb_epoch, it will stop before the nb_epoch epoch is done.
            Default to 100.
        """
        optim = self._handle_batch_size_and_get_optim(batch_size)
        self._aux_var["nb_particles"] = nb_particles
        self._init_proposal_and_particles()
        opt_state = optim.init(self._params)
        nb_epoch_done = 0
        nb_gradients_computed = 0
        nb_max_gradients = np.inf if nb_max_gradients == -1 else nb_max_gradients
        while nb_epoch_done < nb_epoch and nb_gradients_computed < nb_max_gradients:
            loglike = 0
            for selected_indices in self._get_indices_per_batch():
                self._aux_var["selected_indices"] = selected_indices
                self._estimate_proposal()
                self._sample_particles()
                current_loglike, opt_state = self._step(optim, opt_state)
                loglike += current_loglike
                nb_gradients_computed += self._current_batch_size
            self._aux_var["loglike_list"].append(float(loglike / self._nb_batch))
            nb_epoch_done += 1

    def _handle_batch_size_and_get_optim(self, batch_size):
        self._aux_var["batch_size"] = (
            batch_size if batch_size is not None else self.n_samples
        )
        if self._batch_size < self.n_samples:
            optim = optax.adam(learning_rate=0.01)
        else:
            optim = optax.rprop(learning_rate=0.01)
        nb_full_batch = self.n_samples // self._batch_size
        last_batch_size = self.n_samples % self._batch_size
        self._aux_var["nb_batch"] = nb_full_batch + (last_batch_size > 0)
        return optim

    def _get_indices_per_batch(self, shuffle=False):
        indices = np.arange(self.n_samples)
        if shuffle is True:
            np.random.shuffle(indices)

        for i in range(self._nb_batch):
            yield indices[i * self._batch_size : (i + 1) * self._batch_size]

    def _step(self, optim, opt_state):
        (
            normalized_weights,
            const,
            weights,
        ) = self._get_updated_normalized_weights_const_and_weights()
        loglike = self._get_loglike(weights, const)

        grad_params = self._vvmap_grad_theta_log_conditional_y_w(
            self._batch_data,
            self._batch_particles,
            self._params,
            self._batch_additional_data,
        )
        grad_params = self._snis_estimator(grad_params, normalized_weights)
        updates, opt_state = optim.update(grad_params, opt_state)
        self._params = optax.apply_updates(self._params, updates)
        return loglike, opt_state

    def _find_mode(self):
        optim = optax.rprop(0.01)

        def loss(optax_param):
            return -jnp.sum(
                self._vmap_complete_loglike(
                    self._batch_data,
                    optax_param["gaussian"],
                    self._params,
                    self._batch_additional_data,
                )
            )

        optax_params = {
            "gaussian": jnp.zeros((self._current_batch_size, self.latent_dimension))
        }
        opt_state = optim.init(optax_params)
        for _ in range(45):
            grads = grad(loss)(optax_params)
            updates, opt_state = optim.update(grads, opt_state)
            optax_params = optax.apply_updates(optax_params, updates)
        return optax_params["gaussian"]

    def _estimate_proposal(self):
        (
            normalized_weights,
            _,
            _,
        ) = self._get_updated_normalized_weights_const_and_weights()
        expectation_mean = jnp.sum(
            jnp.expand_dims(normalized_weights, 2) * self._batch_particles, axis=0
        )
        self._gaussian_params["mean_prop"] = (
            self._gaussian_params["mean_prop"]
            .at[self._selected_indices]
            .set(expectation_mean)
        )
        best_var = self._get_hessian()
        sqrt_best_var = jnp.linalg.cholesky(best_var)
        self._gaussian_params["var_prop"] = (
            self._gaussian_params["var_prop"].at[self._selected_indices].set(best_var)
        )
        self._gaussian_params["sqrt_var_prop"] = (
            self._gaussian_params["sqrt_var_prop"]
            .at[self._selected_indices]
            .set(sqrt_best_var)
        )

    def _get_hessian(self):
        _hessian = self._vmap_hessian_w_complete_loglike(
            self._batch_data,
            self._batch_mean_prop,
            self._params,
            self._batch_additional_data,
        )
        return jnp.linalg.inv(-_hessian)

    def _get_updated_normalized_weights_const_and_weights(self):
        log_num = self._get_log_numerator()
        log_denom = self._get_log_denominator()
        return _get_normalized_weights_const_and_weights(log_num, log_denom)

    def _get_log_numerator(self):
        return self._vvmap_complete_loglike(
            self._batch_data,
            self._batch_particles,
            self._params,
            self._batch_additional_data,
        )

    def _get_log_denominator(self):
        best_log_density = _vvmap_log_gaussian_density(
            self._batch_particles, self._batch_mean_prop, self._batch_var_prop
        )
        if self._alpha == 0:
            return best_log_density
        batch_identity = jnp.repeat(
            jnp.eye(self.latent_dimension)[None, :, :], self._current_batch_size, axis=0
        ) * (1 + DELTA)
        defensive_log_density = _vvmap_log_gaussian_density(
            self._batch_particles, self._batch_mean_prop, batch_identity
        )
        log_rapport_density = defensive_log_density - best_log_density
        log_rapport_alpha = math.log(self._alpha) + jnp.logaddexp(
            log_rapport_density,
            jnp.array([math.log(1 - self._alpha) - math.log(self._alpha)]),
        )
        return best_log_density + log_rapport_alpha

    def _init_proposal_and_particles(self):
        self._init_proposal()
        self._init_particles()

    def _init_proposal(self):
        self._gaussian_params["mean_prop"] = jnp.zeros(
            (self.n_samples, self.latent_dimension)
        )
        self._gaussian_params["var_prop"] = jnp.repeat(
            jnp.eye(self.latent_dimension)[None, :, :], self.n_samples, axis=0
        )
        self._gaussian_params["sqrt_var_prop"] = jnp.repeat(
            jnp.eye(self.latent_dimension)[None, :, :], self.n_samples, axis=0
        )

    def _sample_particles(self):
        best_one = _sample_gaussians(
            self.nb_particles,
            self._batch_mean_prop,
            self._batch_sqrt_var_prop,
        )
        batch_identity = jnp.repeat(
            jnp.eye(self.latent_dimension)[None, :, :], self._current_batch_size, axis=0
        ) * sqrt(1 + DELTA)
        defensive = _sample_gaussians(
            self.nb_particles, self._batch_mean_prop, batch_identity
        )
        prob = jnp.ones((self.nb_particles, self._current_batch_size)) * self._alpha
        bern = np.random.binomial(1, prob)[:, :, None] * 0
        self._particles = self._particles.at[:, self._selected_indices].set(
            (1 - bern) * best_one + bern * defensive
        )

    def _init_particles(self):
        self._particles = _sample_gaussians(
            self.nb_particles,
            self.mean_prop,
            self._sqrt_var_prop,
        )

    def _init_arguments(self):
        self._aux_var["n_samples"], self._aux_var["dim"] = self._data.shape

    @abstractmethod
    def log_partition(self, x):
        """
        The log partition of the model.
        """

    @abstractmethod
    def link_function(self, x, params, additional_data):
        """
        The link function of the model.
        """

    @abstractmethod
    def log_normalizer(self, x):
        """
        Log normalizer that allows the density to integrate to 1.
        """

    @abstractmethod
    def _get_init_params(self):
        pass



from typing import Optional
import math

import numpy as np
import jax.numpy as jnp
from jax.scipy.stats.binom import logpmf
from jax.scipy.special import logit
from jax.nn import sigmoid
from jax import grad
import torch
from sklearn.decomposition import PCA
import optax


if torch.cuda.is_available():
    DEVICE = "cuda:0"
else:
    DEVICE = "cpu"


def _compute_poissreg_loglike(
    endog: torch.Tensor, offsets: torch.Tensor, exog: torch.Tensor, coef: torch.Tensor
) -> torch.Tensor:
    """
    Compute the log likelihood of a Poisson regression model.

    Parameters
    ----------
    endog : torch.Tensor
        The dependent variable of shape (n_samples, n_features).
    offsets : torch.Tensor
        The offset term of shape (n_samples, n_features).
    exog : torch.Tensor
        The exog of shape (n_samples, n_exog).
    coef : torch.Tensor
        The regression coefficients of shape (n_exog, n_features).

    Returns
    -------
    torch.Tensor
        The log likelihood of the Poisson regression model.

    """
    exog_coef = torch.matmul(exog.unsqueeze(1), coef.unsqueeze(0)).squeeze()
    return torch.sum(
        -torch.exp(offsets + exog_coef) + torch.multiply(endog, offsets + exog_coef)
    )


class _PoissonReg:
    """
    Poisson regression model.

    Attributes
    ----------
    coef : torch.Tensor
        The learned regression coefficients.

    Methods
    -------
    fit(endog, exog, offsets, n_iter_max=300, tol=0.001)
        Fit the Poisson regression model to the given data.

    """

    def __init__(self) -> None:
        """
        Class to infer the maximum likelihood of the coef for the Poisson Regression.
        """
        self._beta: Optional[torch.Tensor] = None

    def fit(
        self,
        endog: torch.Tensor,
        exog: torch.Tensor,
        offsets: torch.Tensor,
    ) -> None:
        """
        Fit the Poisson regression model to the given data.

        Parameters
        ----------
        endog : torch.Tensor
            The dependent variable of shape (n_samples, n_features).
        exog : torch.Tensor
            The exog of shape (n_samples, n_exog).
        offsets : torch.Tensor
            The offset term of shape (n_samples, n_features).
        """
        n_iter_max = 300
        tol = 0.001
        lr = 0.005
        beta = torch.rand(
            (exog.shape[1], endog.shape[1]), requires_grad=True, device=DEVICE
        )
        endog = endog.to(DEVICE)
        exog = exog.to(DEVICE)
        offsets = offsets.to(DEVICE)
        optimizer = torch.optim.Rprop([beta], lr=lr)
        i = 0
        grad_norm = 2 * tol  # Criterion
        while i < n_iter_max and grad_norm > tol:
            loss = -_compute_poissreg_loglike(endog, offsets, exog, beta)
            loss.backward()
            optimizer.step()
            grad_norm = torch.norm(beta.grad)
            beta.grad.zero_()
            i += 1
        self._beta = beta

    @property
    def beta(self):
        """Coef of the Poisson regression."""
        return self._beta


def _init_components_poisson_model(endog: torch.Tensor, rank: int) -> torch.Tensor:
    """
    Initialization for components for the PlnPCA model. Get a first guess for covariance
    that is easier to estimate and then takes the rank largest eigenvectors to get components.

    Parameters
    ----------
    endog : torch.Tensor
        Samples with size (n,p)
    rank : int
        The dimension of the latent space, i.e. the reduced dimension.

    Returns
    -------
    torch.Tensor
        Initialization of components of size (p,rank)
    """
    log_y = torch.log(endog + (endog == 0) * math.exp(-2))
    max_dim = min(rank, endog.shape[0])
    pca = PCA(n_components=max_dim)
    pca.fit(log_y.cpu().detach())
    pca_comp = pca.components_.T * np.sqrt(pca.explained_variance_)
    if rank > max_dim:
        nb_missing = rank - max_dim
        adding = np.random.randn(endog.shape[1], nb_missing) / rank
        pca_comp = np.concatenate((pca_comp, adding), axis=1)
    return pca_comp


def _init_components_binomial_model(
    endog: torch.Tensor, rank: int, n_trials: int
) -> torch.Tensor:
    """
    Initialization for components for the Binomial PCA model. Get a first guess for covariance
    that is easier to estimate and then takes the rank largest eigenvectors to get components.

    Parameters
    ----------
    endog : torch.Tensor
        Samples with size (n,p)
    rank : int
        The dimension of the latent space, i.e. the reduced dimension.
    n_trials : int
        The number of trials assumed in the Binomial model.

    Returns
    -------
    torch.Tensor
        Initialization of components of size (p,rank)
    """
    _endog = endog + (endog == 0) * math.exp(-2) - (endog == n_trials) * math.exp(-2)
    normalized_y = _endog / n_trials
    logit_y = logit(normalized_y)
    max_dim = min(rank, endog.shape[0])
    pca = PCA(n_components=max_dim)
    pca.fit(logit_y)
    pca_comp = pca.components_.T * np.sqrt(pca.explained_variance_)
    if rank > max_dim:
        nb_missing = rank - max_dim
        adding = np.random.randn(endog.shape[1], nb_missing) / rank
        pca_comp = np.concatenate((pca_comp, adding), axis=1)
    return pca_comp


def _compute_binreg_loglike(
    endog: torch.Tensor,
    offsets: torch.Tensor,
    exog: torch.Tensor,
    coef: torch.Tensor,
    n_trials: int,
) -> jnp.ndarray:
    """
    Compute the log likelihood of a Poisson regression model.

    Parameters
    ----------
    endog : torch.Tensor
        The dependent variable of shape (n_samples, n_features).
    offsets : torch.Tensor
        The offset term of shape (n_samples, n_features).
    exog : torch.Tensor
        The exog of shape (n_samples, n_exog).
    coef : torch.Tensor
        The regression coefficients of shape (n_exog, n_features).

    Returns
    -------
    torch.Tensor
        The log likelihood of the Poisson regression model.

    """
    offsets_exog_coef = offsets + jnp.matmul(exog, coef)
    jax_lpmf = logpmf(endog, n_trials, sigmoid(offsets_exog_coef))
    return jnp.sum(jax_lpmf)


class _MBinomialRegression:
    """Multivariate Binomial Regression."""

    def __init__(self, n_trials) -> None:
        self._coef: Optional[jnp.ndarray] = None
        self._n_trials = n_trials

    def fit(
        self,
        endog: jnp.ndarray,
        exog: jnp.ndarray,
        offsets: jnp.ndarray,
    ) -> None:
        """
        Fit the Poisson regression model to the given data.

        Parameters
        ----------
        endog : jnp.ndarray
            The dependent variable of shape (n_samples, n_features).
        exog : jnp.ndarray
            The exog of shape (n_samples, n_exog).
        offsets : jnp.ndarray
            The offset term of shape (n_samples, n_features).

        """
        n_iter_max = 300
        tol = 0.001
        coef = np.random.rand(exog.shape[1], endog.shape[1])
        optim = optax.rprop(learning_rate=0.005)
        dict_coef = {"coef": coef}
        opt_state = optim.init(dict_coef)
        i = 0
        grad_norm = 2 * tol  # Criterion

        def loss(dict_x):
            return -_compute_binreg_loglike(
                endog, offsets, exog, dict_x["coef"], self._n_trials
            )

        while i < n_iter_max and grad_norm > tol:
            grads = grad(loss)(dict_coef)
            updates, opt_state = optim.update(grads, opt_state)
            dict_coef = optax.apply_updates(dict_coef, updates)
            grad_norm = jnp.mean(grads["coef"] ** 2)
            i += 1

        self._coef = coef

    @property
    def coef(self):
        """Coefficient of the Multivariate Binomial regression."""
        return self.coef



# pylint: skip-file

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
from importlib.metadata import version

sys.path.insert(0, os.path.abspath("../../"))

project = "scud"
copyright = "2024, INRAe, Bastien Batardière"
author = "Bastien Batardière"

release = version(project)
version = release

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = ["sphinx.ext.napoleon", "sphinx_rtd_theme"]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]


html_title = "scud"



from abc import ABC, abstractmethod
from typing import Optional, Dict, Union


import numpy as np
import torch
import pandas as pd
import jax.numpy as jnp
from jax import random
from jax.typing import ArrayLike

from scud.inverse_parameter_mappings import poisson_mapping, binomial_mapping
from scud.linkfunctions import linear_with_offsets
from scud._utils import _format_dict_of_array, _add_doc, _vmap_all_but_params


COMPONENT_KEY = "components"
COV_KEY = "exog"
COEF_KEY = "coef"
OFFSETS_KEY = "offsets"


class BaseSampler(ABC):
    """An abstract class used to simulate data from a model."""

    def __init__(
        self,
        n_samples: int,
        latent_dimension: int,
        params: Dict[str, Union[torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]],
        additional_data: Optional[
            Dict[str, Union[torch.Tensor, np.ndarray, pd.DataFrame, jnp.ndarray]]
        ] = None,
    ):
        """
        Instantiate the model with the data given.

        Parameters
        ----------
        n_samples : int
            Number of samples.
        latent_dimension : int
            Latent dimension.
        params : dict
            Model parameters. Each item should be one of
            [torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]
        additional_data : dict, optional
            Additional data, by default None. Each item should be one of
            [torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]

        """
        self._n_samples: int = n_samples
        self._latent_dimension: int = latent_dimension
        self._additional_data: Dict[str, np.ndarray] = _format_dict_of_array(
            additional_data
        )
        self._params: Dict[str, np.ndarray] = _format_dict_of_array(params)

    def sample(self, seed: int = 0) -> jnp.ndarray:
        """
        Generate samples from the model.

        Parameters
        ----------
        seed : int, optional
            Seed for random number generation, by default 0.

        Returns
        -------
        np.ndarray
            Generated samples.
        """
        key = random.PRNGKey(seed)
        gaussian = random.normal(key=key, shape=(self.n_samples, self.latent_dimension))

        parameter = self.vmap_link_function(
            gaussian, params=self.params, additional_data=self._additional_data
        )

        return self.sample_from_parameter(parameter=parameter, key=key)

    @property
    def n_samples(self) -> int:
        """Number of samples."""
        return self._n_samples

    @property
    def latent_dimension(self) -> int:
        """Latent dimension."""
        return self._latent_dimension

    @property
    def params(self) -> Dict[str, np.ndarray]:
        """Method for the parameters of the model."""
        return self._params

    @property
    def additional_data(self) -> Dict[str, jnp.ndarray]:
        """Method for the additional data."""
        return self._additional_data

    @abstractmethod
    def inverse_parameter_mapping(self, x: jnp.ndarray) -> jnp.ndarray:
        """The inverse parameter mapping of the model."""

    @abstractmethod
    def link_function(
        self, x: jnp.ndarray, params: dict, additional_data: dict
    ) -> jnp.ndarray:
        """The link function of the model."""

    @abstractmethod
    def sample_from_parameter(
        self, parameter: jnp.ndarray, key: ArrayLike
    ) -> jnp.ndarray:
        """
        Sample from the distribution of the model given the parameter.

        Parameters
        ----------
        parameter : jax.numpy.ndarray
            Parameter values.
        key : KeyArrayLike
            Random key for sampling.

        Returns
        -------
        jax.numpy.ndarray
            Sampled values.
        """

    def vmap_link_function(
        self, x: jnp.ndarray, params: dict, additional_data: dict
    ) -> jnp.ndarray:
        """
        Vectorized map of the link function.

        Parameters
        ----------
        x : jax.numpy.ndarray
            Input values.
        params : dict
            Model parameters.
        additional_data : dict
            Additional data.

        Returns
        -------
        jax.numpy.ndarray
            Vectorized output values from the link function.
        """
        return _vmap_all_but_params(self.link_function, x, params, additional_data)


class PCAsampler(BaseSampler, ABC):
    """Simulation according to a Principal Component Analysis based model."""

    @classmethod
    def from_dict(cls, dict_params_and_additional_data, **kwargs):
        """
        Create an instance of PCAsampler from a dictionary of parameters and additional data.

        Parameters
        ----------
        dict_params_and_additional_data : dict
            Dictionary containing components, coef, exog, and offsets.
            Each item should be one of
            [torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]
        **kwargs : additional keyword arguments
            Additional arguments for the constructor.

        Raises
        ------
        ValueError
            If the dict does not contain the
            keys "components", "exog", "coef", "offsets".

        Returns
        -------
        PCAsampler
            Instance of PCAsampler created from the provided dictionary.
        """
        for key in [COMPONENT_KEY, COV_KEY, OFFSETS_KEY, COEF_KEY]:
            if key not in dict_params_and_additional_data.keys():
                raise ValueError(f"'{key}' should be in the dictionnary keys.")

        latent_dim = dict_params_and_additional_data[COMPONENT_KEY].shape[1]
        params = {
            COMPONENT_KEY: dict_params_and_additional_data[COMPONENT_KEY],
            COEF_KEY: dict_params_and_additional_data[COEF_KEY],
        }
        additional_data = {
            OFFSETS_KEY: dict_params_and_additional_data[OFFSETS_KEY],
            COV_KEY: dict_params_and_additional_data[COV_KEY],
        }
        n_samples = additional_data[OFFSETS_KEY].shape[0]
        return cls(
            n_samples=n_samples,
            latent_dimension=latent_dim,
            params=params,
            additional_data=additional_data,
            **kwargs,
        )

    @property
    def offsets(self) -> jnp.ndarray:
        """Offsets of the model"""
        return self._additional_data[OFFSETS_KEY]

    @property
    def exog(self) -> jnp.ndarray:
        """Covariates of the model"""
        return self._additional_data[COV_KEY]

    @property
    def components(self) -> jnp.ndarray:
        """Components of the model."""
        return self.params[COMPONENT_KEY]

    @property
    def coef(self) -> jnp.ndarray:
        """Coef of the model."""
        return self.params[COEF_KEY]

    @property
    def covariance(self) -> jnp.ndarray:
        """Covariance matrix of the model."""
        return self.components @ (self.components.T)

    def link_function(
        self,
        x: jnp.ndarray,
        params: Dict[str, Union[torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]],
        additional_data: Dict[
            str, Union[torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]
        ],
    ) -> jnp.ndarray:
        """
        Linear link function.

        Parameters
        ----------
        x : jax.numpy.ndarray
            Input values.
        params : dict
            Model parameters. Each item should be one of
            [torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]
        additional_data : dict
            Additional data. Each item should be one of
            [torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]

        Returns
        -------
        jax.numpy.ndarray
            Output values from the linear link function.
        """
        return linear_with_offsets(x, params=params, additional_data=additional_data)


class PlnPCAsampler(PCAsampler):
    """Simulation according to a Poisson Log-Normal Principal Component Analysis model."""

    def inverse_parameter_mapping(self, x: jnp.ndarray) -> jnp.ndarray:
        """
        Poisson mapping. See wikipédia

        Parameters
        ----------
        x : jax.numpy.ndarray
            Input values.

        Returns
        -------
        jax.numpy.ndarray
            Inverse parameter mapping of the model.
        """
        return poisson_mapping(x)

    @_add_doc(BaseSampler)
    def sample_from_parameter(
        self, parameter: jnp.ndarray, key: ArrayLike
    ) -> jnp.ndarray:
        return random.poisson(key=key, lam=self.inverse_parameter_mapping(parameter))


class BinomialPCAsampler(PCAsampler):
    """Simulation according to a Binomial Principal Component Analysis model."""

    def __init__(
        self,
        n_samples: int,
        latent_dimension: int,
        nb_trials: int,
        params: Dict[str, Union[torch.Tensor, np.ndarray, jnp.ndarray, pd.DataFrame]],
        additional_data: Optional[
            Dict[str, Union[torch.Tensor, np.ndarray, pd.DataFrame, jnp.ndarray]]
        ] = None,
    ):  # pylint: disable=too-many-arguments
        super().__init__(
            n_samples=n_samples,
            latent_dimension=latent_dimension,
            params=params,
            additional_data=additional_data,
        )
        self.nb_trials: int = nb_trials

    @classmethod
    def from_dict(
        cls, dict_params_and_additional_data, nb_trials: int
    ):  # pylint: disable=arguments-differ
        """
        Create an instance of BinomialPCAsampler from a dictionary of
        parameters and additional data.

        Parameters
        ----------
        dict_params_and_additional_data : dict
            Dictionary containing components, coef, exog, and offsets.
        nb_trials : int
            Number of trials for binomial sampling.

        Returns
        -------
        BinomialPCAsampler
            Instance of BinomialPCAsampler created from the provided dictionary.
        """
        return super().from_dict(dict_params_and_additional_data, nb_trials=nb_trials)

    def inverse_parameter_mapping(self, x: np.ndarray) -> np.ndarray:
        """
        Inverse parameter mapping for binomial distribution.

        Parameters
        ----------
        x : jax.numpy.ndarray
            Input values.

        Returns
        -------
        jax.numpy.ndarray
            Inverse parameter mapping of the model.
        """
        return binomial_mapping(x)

    @_add_doc(BaseSampler)
    def sample_from_parameter(
        self, parameter: np.ndarray, key: ArrayLike
    ) -> np.ndarray:
        return random.binomial(
            key=key, n=self.nb_trials, p=self.inverse_parameter_mapping(parameter)
        )
